"""
Solving problems with Dynamic Programming

Copyright 2020, University of Freiburg

Philipp Schneider <philipp.schneider@cs.uni-freiburg.de>
"""

import sys

sys.setrecursionlimit(10000)


def has_equal_partition(array):
    '''
        Returns True if and only if there is a partition of
        the integers in the input array that have equal sum
        of weights.

        Parameter:
            array - contains positive integers

        Returns:
            a boolean value representing whether this array
            can be partitioned into sets with equal sum

        Unit-Tests:
            >>> array = [1]
            >>> has_equal_partition(array)
            False

            >>> array = [1 for i in range(1000)]
            >>> has_equal_partition(array)
            True

            >>> array = [i for i in range(1000)]
            >>> has_equal_partition(array)
            True

            >>> array = [2*i for i in range(1000)] + [1]
            >>> has_equal_partition(array)
            False
    '''
    n = len(array)
    sum = 0

    # compute sum of entries
    for i in range(n):
        sum += array[i]

    # if sum is odd, no solution is possible
    if sum % 2 == 1:
        return False

    # check whether there exists a subset with sum/2
    memo = {}
    return subarray_partition(array, n-1, int(sum/2), memo)


def subarray_partition(array, i, j, memo):
    '''
        Returns True if and only if there is a subset of
        the integers in array[0], ... , array[i] that sums
        up to j

        Parameter:
            array - contains positive integers
            i - integer delimiting the considered part of array
            j - integer that denotes the target sum
            memo - dictionary for story partial results

        Returns:
            a boolean value representing whether  array[0],
            ... , array[i] has a subset that sums up to j

        >>> subarray_partition([1,2,3], 2, 2, {})
        True
        >>> subarray_partition([2,4,6], 2, 3, {})
        False
        >>> subarray_partition([2,4,6], 2, 0, {})
        True
        >>> subarray_partition([2,4,6], 2, -1, {})
        False
        >>> subarray_partition([2,4,6], -1, 2, {})
        False
    '''
    # for j=0 the empty set is a valid solution
    if j == 0:
        return True
    # if the array has just one element it must be equal to j
    if i == 0:
        return array[0] == j
    # no solution possible for empty arrays or negative j
    if j < 0 or i < 0:
        return False

    # check if result is already in memo
    if (i, j) in memo:
        return memo[(i, j)]

    # compute recursive solution
    memo[(i, j)] = subarray_partition(array, i-1, j-array[i], memo)\
        or subarray_partition(array, i-1, j, memo)

    return memo[(i, j)]


def read_array(filename):
    '''
        reads an integer array from a file

        Parameter:
            filename - name of file

        Returns:
            array of integers in the file
    '''
    with open(filename) as input:
        strings = input.read().split(' ')
        array = [int(strings[i]) for i in range(len(strings))]
    return array


def no_consecutive_ones(n):
    '''
        Returns the number of n-digit binary
        strings without consecutive ones

        Parameter:
            n - positive integer

        Returns:
            positive integer

        Unit-Tests:
            >>> no_consecutive_ones(1)
            2

            >>> no_consecutive_ones(3)
            5

            >>> no_consecutive_ones(5)
            13
    '''
    memo = {}
    return no_consecutive_ones_ending(n, 0, memo)\
        + no_consecutive_ones_ending(n, 1, memo)


def no_consecutive_ones_ending(n, i, memo):
    '''
        Returns the number of n-digit binary strings
        ending on i without consecutive ones

        Parameter:
            n - positive integer
            i - 0 or 1
            memo - dictionary for memoization

        Returns:
            positive integer
    '''

    if (n, i) in memo:
        return memo[(n, i)]

    if i == 0:
        if n == 1:
            a = 1
        else:
            a = no_consecutive_ones_ending(n-1, 0, memo)\
                + no_consecutive_ones_ending(n-1, 1, memo)

        memo[(n, i)] = a
        return a

    else:
        if n == 1:
            a = 1
        else:
            a = no_consecutive_ones_ending(n-1, 0, memo)
        memo[(n, i)] = a
        return a
